package store

import (
	"context"
	"math/big"

	"github.com/filecoin-project/lotus/build"
	"github.com/filecoin-project/lotus/chain/state"
	"github.com/filecoin-project/lotus/chain/types"
	big2 "github.com/filecoin-project/specs-actors/actors/abi/big"
	"github.com/filecoin-project/specs-actors/actors/builtin"
	"github.com/filecoin-project/specs-actors/actors/builtin/power"
	cbor "github.com/ipfs/go-ipld-cbor"
	"golang.org/x/xerrors"
)

var zero = types.NewInt(0)

func (cs *ChainStore) Weight(ctx context.Context, ts *types.TipSet) (types.BigInt, error) {
	if ts == nil {
		return types.NewInt(0), nil
	}
	// >>> w[r] <<< + wFunction(totalPowerAtTipset(ts)) * 2^8 + (wFunction(totalPowerAtTipset(ts)) * sum(ts.blocks[].ElectionProof.WinCount) * wRatio_num * 2^8) / (e * wRatio_den)

	var out = new(big.Int).Set(ts.Blocks()[0].ParentWeight.Int)

	// >>> wFunction(totalPowerAtTipset(ts)) * 2^8 <<< + (wFunction(totalPowerAtTipset(ts)) * sum(ts.blocks[].ElectionProof.WinCount) * wRatio_num * 2^8) / (e * wRatio_den)

	tpow := big2.Zero()
	{
		cst := cbor.NewCborStore(cs.Blockstore())
		state, err := state.LoadStateTree(cst, ts.ParentState())
		if err != nil {
			return types.NewInt(0), xerrors.Errorf("load state tree: %w", err)
		}

		act, err := state.GetActor(builtin.StoragePowerActorAddr)
		if err != nil {
			return types.NewInt(0), xerrors.Errorf("get power actor: %w", err)
		}

		var st power.State
		if err := cst.Get(ctx, act.Head, &st); err != nil {
			return types.NewInt(0), xerrors.Errorf("get power actor head: %w", err)
		}
		tpow = st.TotalQualityAdjPower // TODO: REVIEW: Is this correct?
	}

	log2P := int64(0)
	if tpow.GreaterThan(zero) {
		log2P = int64(tpow.BitLen() - 1)
	} else {
		// Not really expect to be here ...
		return types.EmptyInt, xerrors.Errorf("All power in the net is gone. You network might be disconnected, or the net is dead!")
	}

	out.Add(out, big.NewInt(log2P<<8))

	// (wFunction(totalPowerAtTipset(ts)) * sum(ts.blocks[].ElectionProof.WinCount) * wRatio_num * 2^8) / (e * wRatio_den)

	totalJ := int64(0)
	for _, b := range ts.Blocks() {
		totalJ += b.ElectionProof.WinCount
	}

	eWeight := big.NewInt((log2P * build.WRatioNum))
	eWeight = eWeight.Lsh(eWeight, 8)
	eWeight = eWeight.Mul(eWeight, new(big.Int).SetInt64(totalJ))
	eWeight = eWeight.Div(eWeight, big.NewInt(int64(build.BlocksPerEpoch*build.WRatioDen)))

	out = out.Add(out, eWeight)

	return types.BigInt{Int: out}, nil
}
